from typing import List
Vector = List[float]
Matrix2D = List[Vector]

import numpy as np
import matplotlib.pyplot as plt
from AI import mean_squared_error

# Bibliography https://github.com/arseniyturin/SGD-From-Scratch/blob/master/Gradient%20Descent.ipynb

class LinearRegression:

    def fit(self, X: Matrix2D, y: Vector) -> None:
        """Fits the data and calculates the coefficients of the linear regression

        Args:
            X (Matrix2D): Data
            y (Vector): Target
        """
        self.X = np.array(X)
        self.y = np.array(y)
        X, y = self.X, self.y

        if len(X.shape) == 1 or X.shape[1:] == np.ones(X.shape[1:]).all():
            # Least Square Error (minimizes mean square error)
            self.coeffs = ((np.mean(X) * np.mean(y) - np.mean(X*y)) / ((np.mean(X)**2) - np.mean(X**2)))
            self.b = np.mean(y) - self.coeffs * np.mean(X)
            self.uni_dim = True
        else:
            self.coeffs = np.linalg.inv(X.T @ X) @ X.T @ y
            self.b = np.mean(y) - np.mean(X, axis=0) @ self.coeffs
            self.uni_dim = False
    
    def coef_(self) -> Vector:
        """Returns the coefficients

        Returns:
            Vector: The vector of coefficients
        """
        return self.coeffs
    
    def intercept_(self) -> float:
        """Returns the intercept value

        Returns:
            float: Intercept value
        """
        return self.b

    def predict(self, X: Matrix2D = None) -> Vector:
        """Returns the predicted data once fitted

        Args:
            X (Matrix2D, optional): Data to predict. Defaults to None (takes the fitted data)

        Returns:
            Vector: The predicted data
        """
        if X is None: X = self.X

        if self.uni_dim: self.y_pred = X * self.coeffs + self.b
        else: self.y_pred = X @ self.coeffs + self.b
        return self.y_pred
    
    def mse(self, y_real: Vector = None, y_pred: Vector = None) -> float:
        """Returns the mean squared error

        Args:
            y_real (Vector, optional): Real data. Defaults to None (takes the fitted data)
            y_pred (Vector, optional): Predicted data. Defaults to None (takes the predicted data)

        Returns:
            float: Mean squared error
        """
        if y_real is None: y_real = self.y
        if y_pred is None: y_pred = self.y_pred
        return mean_squared_error(y_real, y_pred)

    def plot(self, show: bool = True, delimeters: bool = False) -> None:
        """Plots the linear regression data against the real data

        Args:
            show (bool, optional): This shows the plot. Defaults to True.
            delimeters (bool, optional): This shows the delimeters of the surface that is plot. Defaults to False.
        """
        if self.uni_dim:
            plt.title('Simple Linear Regression')
            plt.ylim(min(self.y), max(self.y))
            plt.plot(self.X, self.y_pred, c='red')
            plt.scatter(self.X, self.y, c='#325aa8', s=15)
            if show: plt.show()

        elif self.X.shape[1] == 2:
            plt.title('Multiple Linear Regression')
            ax = plt.axes(projection = '3d')

            min_x = np.min(self.X, axis = 0)
            max_x = np.max(self.X, axis = 0)

            x_axis = np.array([min_x[0],max_x[0]])
            y_axis = np.array([min_x[1],max_x[1]])

            x1, x2 = np.meshgrid(x_axis, y_axis)
            y = x1 * self.coeffs[0] + x2 * self.coeffs[1] + self.b

            ax.plot_surface(x1, x2, y, color = 'royalblue', alpha = 0.5)
            ax.scatter(self.X[:, 0], self.X[:, 1], self.y, c = 'lightcoral')
            if delimeters: ax.scatter(x1, x2, y, c = 'royalblue', alpha = 0.5)
            if show: plt.show()


class GradientDescent:

    def __init__(self) -> None:
        """Initialize the parameters
        """
        self.m, self.b = 0.2, 0.2
        self.log, self.mse = [], []

    def fit(self, X: Vector, y: Vector, lr: float = 0.05, epoch: int = 10, stochastic: bool = False, batch_size: int = 2) -> None:
        """Calculates the parameters to fit the data

        Args:
            X (Vector): Data
            y (Vector): Target
            lr (float, optional): Describes how big are the steps in every iteration. Defaults to 0.05.
            epoch (int, optional): Describes how many iterations will perform. Defaults to 10.
            stochastic (bool, optional): Performs the sthochastic gradient descend instead. Defaults to False.
            batch_size (int, optional): The size of the data that takes the sthochastic gradient descend. Defaults to 2.
        """
        self.X = np.array(X)
        self.y = np.array(y)
        self.sgd = stochastic
        N, X, y = len(X), self.X, self.y

        for _ in range(epoch):
            if stochastic:
                indexes = np.random.randint(0, len(self.X), batch_size)
                X = np.take(self.X, indexes)
                y = np.take(self.y, indexes)
                N = len(X)
            
            f = y - (self.m * X + self.b)
            # Updating m and b
            self.m -= lr * ((-2 * X @ f).sum() / N)
            self.b -= lr * (-2 * f.sum() / N)
            
            self.log.append((self.m, self.b))
            self.mse.append(mean_squared_error(self.y, (self.m * self.X + self.b)))

    def predict(self, X: Vector = None) -> Vector:
        """Returns the predicted data once fitted

        Args:
            X (Vector, optional): Data to predict. Defaults to None (takes the fitted data)

        Returns:
            Vector: The predicted data
        """
        if X is None: X = self.X
        self.y_pred = self.m * X + self.b
        return self.y_pred

    def plot(self, log: bool = False, show: bool = True, mse: bool = True) -> None:
        """Plots the predicted data against the real data

        Args:
            log (bool, optional): This shows the evolution of the predicted data. Defaults to False.
            show (bool, optional): This shows the plot. Defaults to True.
            mse (bool, optional): This shows the evolution of the error. Defaults to True.
        """
        if mse:
            fig, axs = plt.subplots(2)
            fig.set_size_inches(6.4, 9, forward = True)
            if self.sgd:
                axs[0].set_title('Stochastic Gradient Descent')
                axs[1].set_title('Stochastic Gradient Descent Optimization')
            else:
                axs[0].set_title('Gradient Descent')
                axs[1].set_title('Gradient Descent Optimization')

            axs[1].plot(range(len(self.mse)), self.mse)
            axs[1].set_xlabel('Epochs')
            axs[1].set_ylabel('MSE')

            if log: 
                for i in range(len(self.log)):
                    axs[0].plot(self.X, self.log[i][0] * self.X + self.log[i][1], lw=1, c='orange', alpha=0.25)
            
            axs[0].plot(self.X, self.y_pred, c='red')
            axs[0].scatter(self.X, self.y, c='#325aa8', s=15)
            if show: plt.show()
        
        else:
            if self.sgd:
                plt.title('Stochastic Gradient Descent Optimization')
            else:
                plt.title('Gradient Descent Optimization')
            plt.ylim(min(self.y), max(self.y))
            
            if log: 
                for i in range(len(self.log)):
                    plt.plot(self.X, self.log[i][0] * self.X + self.log[i][1], lw=1, c='orange', alpha=0.25)
            
            plt.plot(self.X, self.y_pred, c='red')
            plt.scatter(self.X, self.y, c='#325aa8', s=15)
            if show: plt.show()

    def plot_mse(self, show: bool = True) -> None:
        """This plots the evolution of the error

        Args:
            show (bool, optional): This shows the plot. Defaults to True.
        """
        plt.title('Gradient Descent Optimization')
        plt.plot(range(len(self.mse)), self.mse)
        plt.xlabel('Epochs')
        plt.ylabel('MSE')
        if show: plt.show()       


def main() -> None:
    """ x = np.random.rand(1000) * 4 -2
    y = np.random.rand(1000) * 4 -2 
    X = np.array([x, y]).T

    Y = x * np.exp(-x**2 - y**2) """
    
    X = np.random.rand(10000) * 4 -2
    Y = X + np.random.random(X.shape)*4

    """ X = np.concatenate((X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X))
    y = np.concatenate((y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y))

    X = np.concatenate((X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X))
    y = np.concatenate((y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y))

    X = np.concatenate((X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X,X))
    Y = np.concatenate((y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y,y)) """

    s = time()
    MLR = LinearRegression()
    MLR.fit(X, Y)
    y_pred = MLR.predict()
    f = time()
    MLR.plot()
    print(f'Error linr: {MLR.mse()}, time:{f-s}')


    s = time()
    GD = GradientDescent()
    GD.fit(X, Y, epoch=50)
    y_pred2 = GD.predict()
    f = time()
    GD.plot(log = True)
    #GD.plot_mse()
    print(f'Error GD  : {GD.mse[-1]}, time:{f-s}')


    s = time()
    SGD = GradientDescent()
    SGD.fit(X, Y, epoch=50, stochastic=True)
    y_pred2 = SGD.predict()
    f = time()
    SGD.plot(log = True)
    #GD.plot_mse()
    print(f'Error SGD : {SGD.mse[-1]}, time:{f-s}')


if __name__ == '__main__':
    from time import time
    main()


import numpy as np
from typing import Callable, List
Function = Callable[[float], float]

def newton(f: Function, df: Function, x: float, tol: float, iter: bool = False) -> float:
    """Newton method to find a root of a function

    Args:
        f (Function): Function
        df (Function): Derivate of the function
        x (float): Start point
        tol (float): Error tolerance
        iter (bool, optional): Shows the iterations needed to find the solution. Defaults to False.

    Returns:
        float: The root
        int: Number of iterations (optional)
    """
    n = 0
    while abs(f(x)) > tol:
        x = x - f(x) / df(x)
        n += 1

    if iter:
        return x, n
    return x

def bisection(f: Function, xi: float, xf: float, tol: float, iter: bool = False) -> float:
    """Bisection method to find a root of a function

    Args:
        f (Function): Function
        xi (float): First point
        xf (float): Second point
        tol (float): Error tolerance
        iter (bool, optional): Shows the iterations needed to find the solution. Defaults to False.

    Returns:
        float: The root
        int: Number of iterations (optional)
    """
    if f(xi) * f(xf) < 0:
        xm, n = (xi + xf) / 2, 1

        while abs(f(xm)) > tol:
            if f(xi) * f(xm) < 0:
                xf = xm 
                n += 1
            
            elif f(xm) * f(xf) < 0:
                xi = xm 
                n += 1
            
            xm = (xi + xf) / 2
            
        if iter:
            return xm, n
        return xm

    else:
        print("Invalid input")

def regula_falsi(f: Function, xi: float, xf: float, tol: float, iter: bool = False) -> float:
    """Regula falsi method to find a root of a function

    Args:
        f (Function): Function
        xi (float): First point
        xf (float): Second point
        tol (float): Error tolerance
        iter (bool, optional): Shows the iterations needed to find the solution. Defaults to False.

    Returns:
        float: The root
        int: Number of iterations (optional)
    """
    if f(xi) * f(xf) < 0:
        xm, n = (xi * f(xf) - xf * f(xi)) / (f(xf) - f(xi)), 1

        while abs(f(xm)) > tol:
            if f(xi) * f(xm) < 0:
                xf = xm
                n += 1

            if f(xm) * f(xf) < 0:
                xi = xm
                n += 1
            
            xm = (xi * f(xf) - xf * f(xi)) / (f(xf) - f(xi))

        if iter:
            return xm, n
        return xm
    else:
        print("Invalid input")


def secant(f: Function, x0: float, x1: float, tol: float, iter: bool = False) -> float:
    """Secant method to find a root of a function

    Args:
        f (Function): Function
        x0 (float): First point
        x1 (float): Second point
        tol (float): Error tolerance
        iter (bool, optional): Shows the iterations needed to find the solution. Defaults to False.

    Returns:
        float: The root
        int: Number of iterations (optional)
    """    
    x2, n = x1 - (f(x1) * (x1 - x0)) / (f(x1) - f(x0)), 1

    while abs(f(x2)) > tol:
        x0, x1 = x1, x2
        x2 = x1 - (f(x1) * (x1 - x0)) / (f(x1) - f(x0))
        n += 1
        
    if iter:
        return x2, n
    return x2



""" def fixed_point(g, dg, x, tol, iter = False):
  ''' f(x) = 0 ==> x = g(x) '''
  n = 1
  if abs(dg(x))<1:
    xa = x
    x = g(x)
    
    while abs(x-xa)>tol:
      xa = x
      x = g(x)
      n = n+1

    if iter: return x, n
    return x

  else:
    print("Doesn't converge")  """


def newton2(f: Function, df: Function, ddf: Function, x: float, tol: float, iter: bool = False) -> float:
    """Newton second order method to find a root of a function

    Args:
        f (Function): Function
        df (Function): Derivate of the function
        ddf (Function): Second derivate of the function
        x (float): Start point
        tol (float): Error tolerance
        iter (bool, optional): Shows the iterations needed to find the solution. Defaults to False.

    Returns:
        float: The root
        int: Number of iterations (optional)
    """
    n = 0

    while abs(f(x)) > tol:
        x1 = x - df(x) / ddf(x) + np.sqrt(df(x)**2 - 2*ddf(x) * f(x)) / ddf(x)
        x2 = x - df(x) / ddf(x) - np.sqrt(df(x)**2 - 2*ddf(x) * f(x)) / ddf(x)

        if abs(f(x1)) < abs(f(x2)):
            x = x1
        else:
            x = x2
        
        n += 1
    
    if iter:
        return x, n
    return x


def main() -> None:
    def f(x): return x**3 + 2*x**2 + 10*x - 20
    def df(x): return 3*x**2 + 4*x + 10
    def ddf(x): return 6*x + 4

    def g(x): return (-x**3 - 2*x**2 + 20) / 10
    def dg(x): return (-3*x**2 - 4*x) / 10

    print('Newton:      ', newton(f, df, 1, 0.01, True))
    print('Bisection:   ', bisection(f, 1, 2, 0.01, True))
    print('Regula falsi:', regula_falsi(f, 1, 2, 0.01, True))
    print('Secant:      ', secant(f, 1, 2, 0.01, True))
    #print(fixed_point(g, dg, 0, 0.1, True))
    print('Newton 2     ', newton2(f, df, ddf, 1, 0.01, True))



if __name__ == '__main__':
    main()



from typing import Callable, List
Vector = List[float]

import numpy as np

def lagrange(X: Vector, Y: Vector) -> Callable[[Vector], Vector]:
    """Returns the lagrange interpolation

    Args:
        X (Vector): X-data
        Y (Vector): Y-data

    Returns:
        Callable[[Vector], Vector]: returns the function that evaluates the lagrange interpolation
    """
    if type(X) == list:
        X = np.array(X)  

    def lagran(x):
        # Checks the input's type
        if type(x) == 'numpy.ndarray':
            x = x.reshape(-1, 1)
        
        if type(x) == list:
            x = np.array(x).reshape(-1, 1)
        
        else:
            x = np.array([x]).reshape(-1, 1)

        out = 0
        for i in range(len(X)):
            #pi_x = (x - np.array([(X[X != xi]).reshape(-1)] * len(x))) because X[X != xi] autoreshape (-1)
            pi_x = x - np.array([(X[X != X[i]])] * len(x))

            out += Y[i] * (pi_x / (X[i] - X[X != X[i]])).prod(axis = 1)

        return out
    
    return lagran


def main() -> None:
    X = np.array([-1, 0, 4,  1, 7, 8])
    X = np.concatenate((X, X+10))
    Y = np.array([ 4, 2, 3, -2, 6, 6])
    Y = np.concatenate((Y, Y+1))

    _x = np.linspace(min(X),max(X),1000)


    Lagran = lagrange(X.reshape(-1, 1),Y.reshape(-1, 1))
    
    #Lagran = lagrange(X,Y)
    
    plt.scatter(X, Y, c='salmon')
    plt.plot(_x, Lagran(_x))
    plt.show()

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    main()


from setuptools import setup, find_packages

VERSION = '0.6'
DESCRIPTION = 'Math and artificial intelligence tools'

# Setting up
setup(
    name="intelligen",
    version=VERSION,
    author="Bouchet07 (Diego Bouchet)",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy', 'matplotlib'],
    keywords=['python', 'AI', 'math', 'artificial', 'intelligence'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)